>* Parsing Text with Nom（【译】使用 Nom 解析文本）
>* 原文链接：https://blog.adamchalmers.com/nom-chars/
>* 原文作者：[Adam Chalmers](https://twitter.com/adam_chal)
>* 译文来自：https://github.com/suhanyujie/article-transfer-rs/
>* ps：水平有限，如有不当之处，欢迎指正。
>* 标签：Rust, Nom, parsing

 “Parsing”是将原始文本或二进制流转换为一些结构化的数据类型，即你的代码可以将其理解和使用的 Rust 类型。这不是这个词的教科书定义，但，这是我的博客中我的见解。本教程是关于 _[nom](https://docs.rs/nom)_（我最喜欢的 Rust 解析库）的。它使用 _parser combinator_ 方法：从编写匹配单个数字或字符的微小解析器开始。这些能成为更大的解析器的构建模块（基础），比如，这些解析器匹配日期或电话号码，通过将许多小型解析器组合在一起，您可以构建一个大型解析器，将文件或流解码为漂亮的 Rust 结构和枚举。在本教程中，我们使用 Nom 将输入文件解析为 [Advent of Code](https://adventofcode.com) 谜题答案。

#  parser 是什么？

大多数程序在知道要做什么之前需要读取一些输入。电子游戏将根据键盘输入移动一个字符，`cat` 命令将根据其参数显示不同的文本，会计软件将根据其加载的电子表格输出一个非常好的或非常令人担忧的信息。

程序输入可以有许多不同的形式。有时候，输入已经是结构化的，例如，大多数电子游戏编程环境可以识别鼠标按钮被按下的时间，知道是哪个鼠标按钮，并调用适当的代码来对此做出反应。但有时输入是非结构化的 —— 它不容易被你的程序理解。例如，输入可以是一个文本文件。您的编程语言可能会理解它得到了一个文本文件。但是它不理解文本的**意思**，也不知道它应该怎么回应。因此，程序员使用解析器来获取这个非结构化文件并检查它，将其与一些逻辑规则进行匹配。一旦数据被结构化，就可以调用适当的函数来处理它。

例如，在 [Advent of Code 第 5 天](https://adventofcode.com/2021/day/5) 中，已知一个行列表，必须找到两行或更多行重叠的点。The Advent of Code 网站提供的谜题输入文本如下:

```
0,9 -> 5,9
8,0 -> 0,8
9,4 -> 3,4
2,2 -> 2,1
7,0 -> 7,4
6,4 -> 2,0
0,9 -> 2,9
3,4 -> 1,4
0,0 -> 8,8
5,5 -> 8,2
```

Rust 并不知道如何将这个文件解释为 2D 空间中的行列表。对于 Rust 编译器来说，这个文本文件只是一个大的 String。我们必须编写一个 _parser_，它接受字符串并将其解析为 Rust 可以理解的类型，比如

```rust
/// 2D 空间中的点（point）
pub struct Point {
    pub x: u32,
    pub y: u32,
}

/// 横跨两点的直线
pub struct Line(pub Point, pub Point);
```

如果您已经了解了 Nom 的基础知识，或者已经在其他语言中使用过解析器组合子，那么您可以[跳过](https://blog.adamchalmers.com/nom-chars/#solve-aoc)，但是现在，让我们进一步了解 Nom。

# 用 Nom 构造解析器

Nom 是一个 _parser combinator_ （解析器组合子）库。一般来说，_combinator_ 框架会给你如下东西：

1.一批小而基础的简单解析器，用来解析简单的文本。

2.组合这些工具以完成更复杂任务的方法。这些叫做组合子。

例如，您可以将 Bash shell 看作一种组合器方法，具有如下特点

1.一组基本的工具，如 ls、cd、cat 和 grep，它们只做一件事，而且做得很好。

2.组合工具的方法，例如 `|`，`&&`，`loops` 和 `if` 语句。

组合子模式的另一个例子是 Rust 的 [Result<T,E>](https://doc.rust-lang.org/stable/std/result/enum.Result.html) 类型。`<T,E>` 你的原始工具可能是类似于 `Fn(T) -> Result<V, U>` 的闭包。组合子就是像 `map`, `and_then` 和 `or_else` 这样的方法。

Nom 采用了类似的方法，使用组合符来组合解析器[1](#1)。有匹配输入字符串中一个简单值的原语解析器，以及允许您将多个原语解析器组合成更复杂的(复合)解析器的函数(称为组合器)。让我们先看看基础的解析器。

## 基础解析器（parser）

解析器的工作是使用来自某个 _input_ 的字节，并尝试将它们转换为某个 _output_（Rust 值）。如果转换成功，解析器应该返回两个东西:

1.未被解析的**剩余输入**，以便您可以继续在输入的剩余部分上运行其他解析器

2.解析的输出值

如果转换失败，解析器应该返回一些有用的错误。

在 Nom 中，解析器只是函数。每个解析器函数都有三种泛型参数:

* `I` 代表 _input_ （输入）。在本教程中，我们的输入将始终是一个字符串，但您可以使用 Nom 来解析二进制(`Vec<u8>` 或 `Bytes`)或您自己的自定义类型，例如由 [lexer](https://en.wikipedia.org/wiki/Lexical_analysis) 输出的令牌。

*  `O` 代表 _output_ （输出）。解析器从输入中读取一些字节，并尝试将它们转换为输出值。一些解析器将输出一个数字、一个字符串或一些“业务逻辑”相关的类型，比如 Person 或 User 结构体。在上面的代码实现中，我们将编写输出 Points 和 Lines 的解析器。

* `E` 是 _error_ （就像标准库中的 `Result<T,E>`）。因为 Nom 支持不同的错误类型的权衡，因而它是通用的。我们以后再讨论。

每个函数的类型签名基本相同: `Fn(I) -> Result<(I, O)， E>`。它只是一个接受输入字符串的函数，读取一些字节，并试图将它们转换为输出值 `O`。如果转换成功，它返回输入的剩余部分（剩余部分，解析器不需要读取来获得值）和值，这就是为什么 Ok 分支下的类型是 `(I, O)`。如果失败，它只返回错误 `E`。

这里有一个非常简单的解析器示例: [`digit1`](https://docs.rs/nom/latest/nom/character/complete/fn.digit1.html)。技术上是 `nom::character::complete::digit1`[2](#2)。nom 文档非常好，我们直接看看他们的例子。

```rust
use nom::character::complete::newline;

fn parser(input: &str) -> IResult<&str, &str> {
    digit1(input)
}

assert_eq!(parser("21c"), Ok(("c", "21")));
assert_eq!(parser("c1"), Err(Err::Error(Error::new("c1", ErrorKind::Digit))));
assert_eq!(parser(""), Err(Err::Error(Error::new("", ErrorKind::Digit))));
```

因此，我们看到 `digit1` 只是一个函数。它接受一个参数，即输入字符串。解析器从输入字符串的开头读取字符，如果字符是数字，则将其添加到输出中。当它发现一个非数字字符时，它终止。一旦它终止，它返回一个 [IResult](https://docs.rs/nom/latest/nom/type.IResult.html)… 那又是什么?

记住上面，我们说过，解析器应该返回 `(I, O)`，如果它们成功（_不是_ 获取输出值的剩余输入字符串，以及输出值本身），如果它们失败，则返回 `E`。`IResult` 是类型的缩写。这是一个方便的[类型别名](https://doc.rust-lang.org/book/ch19-04-advanced-types.html#creating-type-synonyms-with-type-aliases): `type IResult<I, O, E = nom::error::Error<I>> = Result<(I, O), E>` 注意，错误类型 `E` 默认为特定的 Nom 错误类型 —— 我们马上就会讨论这个问题。文档中也很好地解释了 `IResult`:

> 保存解析函数的结果。 `Ok` 时是一个“结果对”，包含输入（未解析的部分数据）的剩余部分和解析的结果。`Err` 端则包含 [`nom::Err`](https://docs.rs/nom/latest/nom/enum.Err.html) 的实例。

正如预期的那样，如果输入字符串 _不_ 以任何数字开头，Nom 将返回一个错误。这个例子使用了 Nom 的默认 Error 类型，它告诉你两件事:

1.入参有问题导致 Nom 解析失败。
2.Nom 尝试解析失败了

如果你想要更多的细节，你可以使用 `E` = [`VerboseError`](https://docs.rs/nom/latest/nom/error/struct.VerboseError.html) 来代替默认的错误类型。这给运行时增加了一点开销，但使调试更容易。

让我们看看另一个名为 [`tag`](https://docs.rs/nom/latest/nom/bytes/complete/fn.tag.html) 的解析器。它匹配由程序员选择的特定子字符串。Nom 文档中有一个例子，使用 `tag` 来匹配子字符串 “Hello”：

```rust
use nom::bytes::complete::tag;

fn parser(s: &str) -> IResult<&str, &str> {
  tag("Hello")(s)
}

assert_eq!(parser("Hello, World!"), Ok((", World!", "Hello")));
assert_eq!(parser("Something"), Err(Err::Error(Error::new("Something", ErrorKind::Tag))));
assert_eq!(parser(""), Err(Err::Error(Error::new("", ErrorKind::Tag))));
```

注意，`tag` 函数与 `digit1` 函数有不同的类型签名！这是因为 `digit1` 是一个解析器 —— 它接受一个输入字符串并输出一个 `IResult`，定义为["解析函数的结果"](https://docs.rs/nom/latest/nom/type.IResult.html)。但是 `tag` 并不是一个解析器 —— 它是一个 _输出_ 为解析器的函数。它接受一个参数，即要匹配的子字符串，并返回匹配该字符串的解析器。

*  `digit` 是一个解析器，它接受一个参数（一个输入字符串）并返回 `IResult`。
*  `tag` 接受一个参数，一个要匹配的字符串，并返回一个解析器。该解析器接受一个参数（一个输入字符串）并返回 `IResult`。

Nom 的其他基本解析器可以分为以下几类：它们要么是真正的解析器，要么是返回解析器的函数。例如:

* [`digit0`](https://docs.rs/nom/latest/nom/character/complete/fn.digit0.html)，[`alpha1`](https://docs.rs/nom/latest/nom/character/complete/fn.alpha1.html) 和 [`line_ending`](https://docs.rs/nom/latest/nom/character/complete/fn.line_ending.html) 是解析器
* [`char`](https://docs.rs/nom/latest/nom/character/complete/fn.char.html) 接受一个字符，并返回匹配该字符的解析器。

好了，现在我们可以构建小型的解析器来识别特定的字符。但是这还不足以解析 Advent of Code 输入文件。我们需要一个解析器来解析文本“(3,4)”为 `Point{x: 3, y: 4}`。为此，我们需要将这些小型解析器组合到一个复合解析器中，该复合解析器可以将文本转换为 Point。

## 用。。。组合解析器 组合子

到目前为止，我们已经看到了两种类型的 Nom 函数:

1.解析器，如：`digit1`
2.返回解析器的函数，如：`tag`

我们将研究第三种:

3.接受一个解析器作为参数并返回一个新的解析器的函数，例如 [`pair`](https://docs.rs/nom/latest/nom/sequence/fn.pair.html)。

 `pair` 函数以两个解析器作为参数（分别称为 `p1` 和 `p2`）。它返回一个新的解析器，将它们链接在一起。每个解析器接受它的输入类型 I （通常是字符串），并输出 (I, O) 作为剩余的输入和它解析的输出值：

![the input remaining after p1 becomes the input to p2, and their outputs get paired together](/nom-chars/parser_pair_railroad.png)

大致流程如下：

1.执行 `p1`
2.如果 `p1` 解析失败，返回错误
3.否则，`p1` 执行后的“剩余输入”将成为 `p2` 的输入
4.执行 `p2`

通常，官方文档中有一个很好的示例：

```rust
use nom::sequence::pair;
use nom::bytes::complete::tag;

// `pair` 从第一个解析器获取一个对象，然后从第二个解析器获取另一个对象。
let mut parser = pair(tag("abc"), tag("efg"));

assert_eq!(parser("abcefg"), Ok(("", ("abc", "efg"))));
assert_eq!(parser("abcefghij"), Ok(("hij", ("abc", "efg"))));
assert_eq!(parser(""), Err(Err::Error(("", ErrorKind::Tag))));
assert_eq!(parser("123"), Err(Err::Error(("123", ErrorKind::Tag))));
```

这是一个非常简单的解析器组合子它接受两个解析器并输出一个新的解析器。Nom 有很多这些有用的小功能。例如，有 [`separated_pair`](https://docs.rs/nom/latest/nom/sequence/fn.separated_pair.html)，它接受 _three_ 解析器：它依次运行所有三个解析器，并丢弃中间解析器的输出。它对删除分隔符很有用，例如将“1,2”解析为“1”和“2”。

```rust
fn main() {
    use nom::{character::complete::{char, digit1}, sequence::separated_pair};

    let mut parser = separated_pair(digit1, char(','), digit1);
    assert_eq!(parser("1,2"), Ok(("", ("1", "2"))));
}
```

到目前为止，我们已经看到了接受多个解析器作为参数的组合子。但是这里有一个只接受一个解析器的方法：它被称为 [`map`](https://docs.rs/nom/latest/nom/combinator/fn.map.html)。它包装一个解析器，然后将输出传递到一个闭包中。

```rust
use nom::{
    character::complete::{alpha1, char},
    combinator::map,
    sequence::separated_pair,
    IResult,
};

/// 程序使用的一些业务逻辑示例类型
#[derive(PartialEq, Eq, Debug)]
struct Person {
    first_name: String,
    last_name: String,
}

impl Person {
    fn parse(input: &str) -> IResult<&str, Self> {
        // 解析用空格分隔的两个单词。
        // 注意：alpha1 可以识别从 A 到 Z 的一个或多个字母。
        let two_words_parser = separated_pair(alpha1, char(' '), alpha1);

        // 定义一个新的解析器来包装 `two_words_parser`，然后
        // 将“结果对”传递到闭包中。
        let mut person_parser = map(
            two_words_parser,
            |(first_name, last_name)| Self {
                first_name: String::from(first_name),
                last_name: String::from(last_name),
            },
        );

        // 使用解析器
        person_parser(input)
    }
}

fn main() {
    let (remaining_input, actual) = Person::parse("Adam Chalmers_").unwrap();
    let expected = Person {
        first_name: "Adam".to_owned(),
        last_name: "Chalmers".to_owned(),
    };
    assert_eq!(remaining_input, "_");
    assert_eq!(expected, actual);
}
```

`map` 组合子入参是可以执行成功的闭包。但有时你想要处理一个可能失败的结果。例如，您可能有一个解析器输出一个字符串，然后尝试将该字符串转换为数字。如果闭包返回一个 Result，我们可以使用 `map_res`。

```rust
use nom::{character::complete::digit1, combinator::map_res, IResult};
use std::str::FromStr;

/// 从字符串解析一个数字，并返回一个实际的 Rust 数字，而不是一个字符串。
pub fn parse_numbers(input: &str) -> IResult<&str, u32> {
    let mut parser = map_res(digit1, u32::from_str);
    parser(input)
}

fn main() {
    let (actual_input_remaining, actual_output) = parse_numbers("123").unwrap();
    let expected_output = 123;
    let expected_input_remaining = "";
    assert_eq!(actual_output, expected_output);
    assert_eq!(actual_input_remaining, expected_input_remaining);
}
```

Nom 模块中还有很多其他组合子：

* [`nom::sequence`](https://docs.rs/nom/latest/nom/sequence/index.html) 中包含我们已经见过的 `pair` 和 `separated_pair` 组合子。其他还有 [`tuple`](https://docs.rs/nom/latest/nom/sequence/fn.tuple.html)，它接受任意数量的不同解析器并执行。`pair` 是 `tuple` 的一个特例，是长度为 2 的元组。
* [`nom::multi`](https://docs.rs/nom/latest/nom/multi/index.html) 中有接收一个解析器作为参数，并能执行多次的组合子，例如 [`many0`](https://docs.rs/nom/latest/nom/multi/fn.many0.html) 包装另一个解析器并执行 0 次或多次， 还有 [`separated_list1`](https://docs.rs/nom/latest/nom/multi/fn.separated_list1.html) 可以将 `"1,2,3"` 解析为 `vec![1,2,3]`
* [`nom::branch`](https://docs.rs/nom/latest/nom/branch/index.html) 中有类似于 [`alt`](https://docs.rs/nom/latest/nom/branch/fn.alt.html) 的组合子，用于尝试几种不同的解析器，只需之一能成功执行即可。（例如 将 "3" 或 0003 解析为数字 3）
* [`nom::combinator`](https://docs.rs/nom/latest/nom/combinator/index.html) 中有很多通用的组合子，例如 [`map_opt`](https://docs.rs/nom/latest/nom/combinator/fn.map_opt.html)（类似于 `map_res`，但闭包返回 Option，而非 Result）

好了！这就是 Nom 的基础：解析器使用来自输入（通常是字符串）的字节，并输出一个 Rust 值。有一些匹配简单模式的原始解析器，比如一些字母数字字符或空白字符解析器。然后是组合子，它接受一个或多个解析器，并产生更复杂的解析器。然后，这些解析器可以成为更多组合子的输入，从而形成更复杂的解析器。如果你一直使用这个过程，就可以解析一些非常复杂的格式，比如 HTTP 或 EXIF。让我们看看我使用 Nom 的一个真实例子:解析 Advent of Code 第 5 天的代码。

# 用 Nom 解决问题

上面我们[描述了这个问题](https://blog.adamchalmers.com/nom-chars/#what-are-parsers)：读取一个文本文件并将其解析为 `Line` 对象。使用 Nom 进行解析的关键是**从小部分**开始。找到你知道解析的一小部分输入，并对其进行单元测试，以确保其工作正常。然后在此基础上构建更复杂的解析器。

显然，我们需要知道如何解析单个 2D 点。

```rust
use nom::{
    character::complete::{char, digit1},
    combinator::{map, map_res},
    sequence::separated_pair,
    IResult,
};
use std::str::FromStr;

/// 从输入的字符串开始部分解析 `u32`。
/// （这是从早些时候的博客文章中复制/粘贴出来的！）
pub fn parse_numbers(input: &str) -> IResult<&str, u32> {
    map_res(digit1, u32::from_str)(input)
}

/// A point in 2D space
#[derive(Debug, Eq, PartialEq)]
pub struct Point {
    pub x: u32,
    pub y: u32,
}

impl Point {
    /// 解析一个点
    /// 像所有的 Nom 解析器一样，它有 Input 类型，Output 类型和 Error 类型。
    /// 它的输入类型是 &str (因为我们解析的是一个文本文件，将内容读取到一个字符串)
    /// 它的输出类型是 Point (我们可以在 `impl Point` 块中称其为 Self)
    /// 它的 Error 类型将是默认的 Nom 错误类型，所以我们不必明确写出
    fn parse(input: &str) -> IResult<&str, Self> {

        // 解析器输出一个 (u32, u32)。
        // 它使用 `parse_numbers` 解析器
        // 和前面讨论过的 `separated_pair` 组合符。
        let parse_two_numbers = separated_pair(parse_numbers, char(','), parse_numbers);

        // 将 (u32, u32) 映射到一个 Point。
        // 使用上一行定义的 `parse_two_numbers` 解析器，
        // 和前面讨论过的 `map` 组合子
        map(parse_two_numbers, |(x, y)| Point { x, y })(input)
    }
}

// 因为每个解析器只是一个纯粹的、确定性的函数，所以对它们进行单元测试非常容易！
#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_parse_point() {
        let tests = [
            ("1,2", Point { x: 1, y: 2 }, ""),
            ("1,2asdf", Point { x: 1, y: 2 }, "asdf"),
        ];
        for (input, expected_output, expected_remaining_input) in tests {
            let (remaining_input, output) = Point::parse(input).unwrap();
            assert_eq!(output, expected_output);
            assert_eq!(remaining_input, expected_remaining_input);
        }
    }
}
```

好了,好了。现在我们可以将 Point 解析器用于 Line 解析。Advent of Code 中的文本文件按这样的方式存储行：

> 0,9 -> 5,9

```rust
/// 横跨两点的直线
#[derive(Debug, Eq, PartialEq)]
pub struct Line(pub Point, pub Point);

impl Line {
    /// 根据输入解析为 Line
    fn parse(input: &str) -> IResult<&str, Self> {
        let parse_arrow = tag(" -> ");

        // 解析由箭头分隔的两个 Point
        let parse_points = separated_pair(Point::parse, parse_arrow, Point::parse);

        // 若解析成功，则将这两个 point 放到 Line 中
        map(parse_points, |(p0, p1)| Line(p0, p1))(input)
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_parse_line() {
        let tests = [
            (
                "0,9 -> 5,9",
                Line(Point { x: 0, y: 9 }, Point { x: 5, y: 9 }),
                "",
            ),
            (
                "0,9 -> 5,9xyz",
                Line(Point { x: 0, y: 9 }, Point { x: 5, y: 9 }),
                "xyz",
            ),
        ];
        for (input, expected_output, expected_remaining_input) in tests {
            let (remaining_input, output) = Line::parse(input).unwrap();
            assert_eq!(remaining_input, expected_remaining_input);
            assert_eq!(output, expected_output);
        }
    }
}
```

现在解析剩下的部分应该很容易了。我们知道文本文件中的每一行都应该编码为 Line 对象。我们可以使用 [`separated_list1`](https://docs.rs/nom/latest/nom/multi/fn.separated_list1.html) 来匹配一行，然后匹配并丢弃换行符，然后匹配一行，然后匹配并丢弃换行符，以此类推，直到文件结束。

```rust
use nom::character::complete::line_ending;

/// 解析整个 Advent of Code 第5天的文本文件。
pub fn parse_input(s: &str) -> Vec<Line> {
    let (remaining_input, lines) = separated_list1(line_ending, Line::parse)(s).unwrap();
    assert!(remaining_input.is_empty());
    lines
}


#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_parse_file() {
        // Assuming you're logged in, you can download your Day 5 puzzle input from
        // https://adventofcode.com/2021/day/5/input
        // Then save it under src/data/input.txt
        // The `include_str!` macro reads a file into a String at compile-time.
        // I use it for all my AoC problems!
        let input = include_str!("data/input.txt");
        let lines = parse_input(input);
        assert_eq!(lines.len(), 500);
    }
}
```

现在我们可以解析整个文本文件！你可以在 [GitHub](https://github.com/adamchalmers/aoc21/blob/master/day5/src/parse.rs) 上看到完整的代码，如果你有任何建议，欢迎开启 PR。

我发现“从小做起”的方法真的很有成效。使用许多小的解析器而不是一个大的解析器，可以很容易地理解和测试每一个小步骤。我可以在这里选择任何解析器，并通过查看类型签名确切地了解它是做什么的。如果我想知道它解析的确切文本类型，我可以检查单元测试。这样即使在我几年后，也很容易理解它。我应该知道这一点，因为几年前我用 Nom 编写了[fountain crate](https://crates.io/crates/fountain)，它可以解析[Fountain markdown 语言](https://fountain.io/)，我仍然发现代码[相当可读](https://github.com/adamchalmers/fountain-rs/blob/master/fountain/src/parse.rs)[3](#3)。

# 下一步

我希望这能帮助你理解 Nom 和解析器组合子！如果你想了解更多，我建议你看看 Nom 的 [README.md](https://github.com/Geal/nom/blob/main/README.md) —— 它有一些不同文档的链接，包括 API 文档，一些设计文档，甚至还有一个“我应该使用哪个组合符”指南。Nom 的创造者 Geoffroy Couprie 制作的文档真的已经超越很多其他项目文档！

我们将在以后的文章中进一步介绍 Nom。现在，如果你有任何问题，请在下方留言或在 twitter 上与我交谈 [@adam_chal](https://twitter.com/adam_chal)。如果你喜欢这篇文章，想和我一起全职写 Rust 并获得报酬，请将你的简历发送至 ehfg-wbof@pybhqsyner.pbz rot13。上一个这样联系的人被录用了，所以不要害羞!我们目前正在欧盟和美国招聘远程工作者（我们在[很多城市](https://www.cloudflare.com/careers/jobs/)都有办事处）。

_感谢 Joshua Nelson, Michael Sproul 和 Stephen Merity 阅读草案！_
* * *

#### 脚注

1.Nom 的灵感来自于 Haskell 的原始解析器组合子库 Parsec。我第一次了解 Parsec 是在 [Real World Haskell](http://book.realworldhaskell.org/read/using-parsec.html) 的章节中，这个概念可以如此轻松地移植到 Rust 中，这让我印象深刻。

2.有两种解析器，_complete_ 用于将整个输入读入内存，_streaming_ 用于逐位获取输入。对于  Advent of Code，输入文件总是足够大，以适应内存。但是如果是基于流的输入，比如说，非常大的数据集无法放入内存，你会使用 _streaming_。或者，如果您从流协议（如 HTTP 或 gRPC）中读取数据，你可能也会在那个地方使用 streaming。

3.注意类型有很多复杂的模板 —— 这是因为我想让解析器支持默认的 Nom 错误或 Nom 的 VerboseError。如果你有改进它的建议而不破坏向后兼容性，请让我知道！
