>* Ownership and impl Trait 译文（Rust 中的所有权和 impl trait）
>* 原文链接：https://www.fpcomplete.com/rust/ownership-and-impl-trait/
>* 原文作者：[FPComplete](https://www.fpcomplete.com/)
>* 译文来自：https://github.com/suhanyujie/article-transfer-rs/
>* 译者：[suhanyujie](https://ishenghuo.cnblogs.com/)
>* ps：水平有限，如有不当之处，欢迎指正
>* 标签：Rust，onwership，impl

Rust API 中有一个常见的模式: 返回一个相对复杂的数据类型，且该类型提供了我们想要使用的 trait 实现。许多 Rust 新手首先遇到的类似对象就是迭代器。例如，如果我想提供一个返回数字 1 到 10 范围的函数，它可能看起来像这样:

```rust
use std::ops::RangeInclusive;

fn one_to_ten() -> RangeInclusive<i32> {
    1..=10i32
}
```

这模糊所发生的迭代特性。因而，当需求变得更复杂时，情况会变得更糟，例如:

```rust
use std::iter::Filter;

fn is_even(x: &i32) -> bool {
    x % 2 == 0
}

fn evens() -> Filter<RangeInclusive<i32>, for<'r> fn(&'r i32) -> bool> {
    one_to_ten().filter(is_even)
}
```

再复杂一点：

```rust
use std::iter::Map;

fn double(x: i32) -> i32 {
    x * 2
}

fn doubled() ->
    Map<
        Filter<
               RangeInclusive<i32>,
               for<'r> fn(&'r i32) -> bool
              >,
        fn(i32) -> i32
       > {
    evens().map(double)
}
```

这显然不是我们想要编写的代码！幸运的是，我们现在有了一种更优雅的方式来表达我们的意图: `impl Trait`。这个特性允许我们表达一个函数返回一个值，且该值实现了某个 trait，而非显式地声明具体的类型。我们可以用以下方法重写上面的签名:

```rust
fn one_to_ten() -> impl Iterator<Item = i32> {
    1..=10i32
}

fn is_even(x: &i32) -> bool {
    x % 2 == 0
}

fn evens() -> impl Iterator<Item = i32> {
    one_to_ten().filter(is_even)
}

fn double(x: i32) -> i32 {
    x * 2
}

fn doubled() -> impl Iterator<Item = i32> {
    evens().map(double)
}

fn main() {
    for x in doubled() {
        println!("{}", x);
    }
}
```

这对于开发来说是一个福音，特别是当我们遇到更复杂的情况时(比如 future 和 庞杂的 Tokio 代码)。然而，我想展示一个 impl Trait 示例。希望这将有助于解释所有权的一些细微差别及其与该特性的交互。

## 引入一个问题
看一下这段不能编译的代码:

```rust
// 试试用 (_: &String) 替换
fn make_debug<T>(_: T) -> impl std::fmt::Debug {
    42u8
}

fn test() -> impl std::fmt::Debug {
    let value = "value".to_string();

    // 尝试移除 `&` 以通过编译
    make_debug(&value)
}

pub fn main() {
    println!("{:?}", test());
}
```

在这段代码中，有一个 `make_debug` 函数，它接受任何值，完全忽略该值，并返回一个 `u8`。但是，在函数签名中并未包含 `u8`，而是 `impl Debug` (这是完全有效的: `u8` 实际上实现了 `Debug`)。`test` 函数通过向 `make_debug` 传入 `&String` 来返回 `impl Debug`。

当我尝试编译它时，我得到了错误消息:

```
error[E0597]: `value` does not live long enough
  --> src/main.rs:10:16
   |
6  | fn test() -> impl std::fmt::Debug {
   |              -------------------- opaque type requires that `value` is borrowed for `'static`
...
10 |     make_debug(&value)
   |                ^^^^^^ borrowed value does not live long enough
11 | }
   | - `value` dropped here while still borrowed
```

在我们试图理解这个错误消息之前，我想在这里强调该问题点。我可以对这段代码进行一些更改，以使其通过编译。例如:

* 如将 `make_debug` 中的 `T` 参数替换为 `&String`（或者更常用的 `&str`）。由于某种原因，多态性会导致一些问题。
* 把签名由 `make_debug<T>(_: T)` 改为 `make_debug<T>(_: &T)` 也能修复，但可能会更奇怪。
* 最后，在调用 `make_debug` 后，如果入参是值本身，而非值的引用，就能编译了。即：`make_debug(value)` 而非 `make_debug(&value)`。直觉上，希望使用引用时，不要遇到生命周期错误。

一些事情正在变的微妙，我们试着慢慢理解它。

## 具体类型的生命周期
让我们将 `make_debug` 函数简化为显式接受一个 `String`:

```rust
fn make_debug(_: String) -> impl std::fmt::Debug {
    42u8
}
```

这时，参数的生命周期是多少？`make_debug` 接收该值作为入参，然后丢弃它。该值不能再在函数之外使用。但有趣的是，`make_debug` 丢弃它的事实并没有反映在函数类型签名中; 它只表明返回一个 `impl Debug`。为了证明这一点，我们可以返回参数本身，而不是我们的 `42u8`：

```rust
fn make_debug(message: String) -> impl std::fmt::Debug {
    //42u8
    message
}
```

在这种情况下，`message` 的所有权从 `make_debug` 函数本身转移到返回的 `impl Debug` 值上。这是一个有趣而重要的现象，我们一会儿再回来。让我们继续探索，看看接受 `&String` 版本的 `make_debug`：

```rust
fn make_debug(_: &String) -> impl std::fmt::Debug {
    42u8
}
```

那个引用的生命周期是多久？由于生命周期参数省略了，我们不必明确地说明它。但是隐含的生命周期在函数本身的生存期内。换句话说，当函数退出时，`String` 的借用则过期。我们可以通过返回引用来证明这一点：

```rust
fn make_debug(message: &String) -> impl std::fmt::Debug {
    //42u8
    message
}
```

我们得到的错误消息有点出人意料，但相当有帮助:

```rust
error: cannot infer an appropriate lifetime
 --> src/main.rs:4:5
  |
2 | fn make_debug(message: &String) -> impl std::fmt::Debug {
  |                                    -------------------- this return type evaluates to the `'static` lifetime...
3 |     //42u8
4 |     message
  |     ^^^^^^^ ...but this borrow...
  |
note: ...can't outlive the anonymous lifetime #1 defined on the function body at 2:1
 --> src/main.rs:2:1
  |
2 | / fn make_debug(message: &String) -> impl std::fmt::Debug {
3 | |     //42u8
4 | |     message
5 | | }
  | |_^
help: you can add a constraint to the return type to make it last less than `'static` and match the anonymous lifetime #1 defined on the function body at 2:1
  |
2 | fn make_debug(message: &String) -> impl std::fmt::Debug + '_ {
  |                                    ^^^^^^^^^^^^^^^^^^^^^^^^^
```

现在的情况是，我们的签名本质上有两个生命周期。`message` 的隐含生命周期就是函数的生命周期，而 `impl Debug` 的生命周期是 `'static`，这意味着它要么不借用数据，要么只借用持续整个程序的值(比如字符串文字)。我们甚至可以尝试贯彻这个建议，并添加一些明确的生命周期：

```rust
fn make_debug<'a>(message: &'a String) -> impl std::fmt::Debug + 'a {
    message
}

fn test() -> impl std::fmt::Debug {
    let value = "value".to_string();
    make_debug(&value)
}
```

虽然这个修复程序 `make_debug` 本身，但是我们不能再通过 `test` 调用 `make_debug`:

```
error[E0597]: `value` does not live long enough
  --> src/main.rs:11:16
   |
7  | fn test() -> impl std::fmt::Debug {
   |              -------------------- opaque type requires that `value` is borrowed for `'static`
...
11 |     make_debug(&value)
   |                ^^^^^^ borrowed value does not live long enough
12 | }
   | - `value` dropped here while still borrowed
```

换句话说，我们假设 `test()` 的返回值比 `test` 本身存在时间长，但是 `value` 不会比 `test` 存在时间长。

**思考问题**确保你可以向你自己(或者一只橡皮鸭)解释: 为什么返回的 `message` 可以通过值传递而不是通过引用传递呢？

对于 `make_debug` 的具体类型版本，我们基本上有一个二乘二的矩阵: 是通过值传递还是通过引用传递，以及是返回提供的参数还是虚拟的 `42u8` 值。让我们清楚地记录下来:

|        | 值传递| 引用传递 |
| ----------- | ----------- |---|
| Use message |  Success: 返回值拥有所有权 | 失败: 返回值超过引用的生存期  |
| Use dummy |  Success: 返回值无需参数 | 失败: 返回值无需引用  |

## 为什么多态会破坏 Rust 机制？
在下面的行中可以看到，当返回 42 时，使用值传递和引用传递都是安全的，因为返回的值根本不需要生命周期参数。但是由于某种原因，当我们使用参数 T 而不是 `String` 或 `&String` 时，我们会得到一个错误消息。让我们用代码来回忆一下：

```rust
fn make_debug<T>(_: T) -> impl std::fmt::Debug {
    42u8
}

fn test() -> impl std::fmt::Debug {
    let value = "value".to_string();
    make_debug(&value)
}
```

错误消息：

```
error[E0597]: `value` does not live long enough
  --> src/main.rs:10:16
   |
6  | fn test() -> impl std::fmt::Debug {
   |              -------------------- opaque type requires that `value` is borrowed for `'static`
...
10 |     make_debug(&value)
   |                ^^^^^^ borrowed value does not live long enough
11 | }
   | - `value` dropped here while still borrowed
```

从 `make_debug` 内部，我们可以很容易地看到参数被忽略。但是，这一点很重要: `make_debug` 的函数签名并没有明确地告诉我们这一点！相反，我们知道的是:

* `make_debug` 携带了类型参数 T
* T 可能隐含了非静态生命周期的引用参数，我们确实无法确定（这一点很重要）
* 返回值是 `impl Debug`
* 我们不知道有哪些具体的类型满足返回值的约束，但一定符合 `'static` 生命周期
* `impl Debug` 可能由上层的 T 参数而定

这样做的结果是: 如果 T 有任何引用，那么它们的生存期必须至少与返回 `impl Debug` 的生存期一样久，这意味着它必须是 `'static` 生命周期。这就是我们得到的错误消息:

```
opaque type requires that `value` is borrowed for `'static`
```

请注意，这发生在 `make_debug` 调用中，而不是 `make_debug` 内部。我们的 `make_debug` 函数是完全有效的，它只是有一个隐含的生命周期。我们可以确切地表示出来:

```rust
fn make_debug<T: 'static>(_: T) -> impl std::fmt::Debug + 'static
```

## 为什么这些改变能奏效

我们以前通过使参数的类型具体化来修复编译失败。有两种相对简单的方法可以解决编译失败并保持类型多态。他们是:

* 将参数由 `_: T` 改为 `_: &T`
* 调用处由 `make_debug(&value)` 改为 `make_debug(value)`

挑战，在阅读下面的解释之前，根据我们到目前为止所解释的内容，尝试弄清楚这些更改为何能修复编译。

## 将参数更改为 `&T`
我们对 T 的隐式要求是，它包含的任何引用都具有 static 生命周期。这是因为我们无法从类型签名中看出 `impl Debug` 是否保存了 `T` 中的数据。然而，通过使参数本身成为引用，我们完全改变了游戏规则。突然之间，我们有了两个隐含生命周期参数，而不仅仅 `T` 上静态的隐含生命周期参数:

* 引用的生命周期参数，我们成为 `'a`
* T 和 `impl Debug` 的生命周期都是 `'static`

表示成如下：

```rust
fn make_debug<'a, T: 'static>(_: &'a T) -> impl std::fmt::Debug + 'static
```

虽然我们不能从这个类型签名看出 `impl Debug` 是否依赖于 `T`，但是我们确实知道ーー通过 `impl Trait` 特性本身的定义ーー它不依赖于 `'a` 生命周期。因此，对引用的唯一要求是，只要 `make_debug` 调用本身一直存在，这就是对的。

## 将调用更改为值传递（pass-by-value）
另一方面，如果将参数变为 `T` (而不是 `&T`) ，则可以通过使用 make_debug(value)（而不是 make_debug(&value)）传递 value 来修复编译问题。这是因为传入的 T 值的要求是它具有`'static` 生命周期，而没有引用的值确实具有这样的生存期(因为它们属于函数所有)。更直观地说: `make_debug` 具有 T 的所有权，如果返回值由 `impl Debug` 改为 `T`，它将从 `make_debug` 获得 `T` 的所有权。否则，当离开 `make_debug` 时，`T` 将被回收。

## 按表格查看
为了总结多态情况，让我们列出另一个表，这次比较的参数是 `T` 还是 `&T`，以及调用是 `make_debug(value)` 还是 `make_debug(&value)`:

|        | 值传递：`T`| 引用传递: `&T` |
| ----------- | ----------- |---|
| `make_debug(value)` | 成功：String 的生命周期是 `'static`   | 类型错误：当期望参数是引用时，传入 `String`  |
| `make_debug(&value)` | 生命周期错误：`&String` 并没有 `'static` 那么久 | 成功：生命周期参数是 `'a`   |

## 结论
就我个人而言，我发现 `impl Trait` 这种行为最初令人困惑。然而，通过上面的步骤帮助我更好地理解了这种情况下的所有权。`impl Trait` 是 Rust 语言的一个重要特征。但是，在某些情况下，我们可能需要更明确的生存期，然后可能需要补全为完整类型签名方法。希望这样的案子少之又少。通常，明确的 `clone` (尽管效率低下)可以节省大量工作。

## 学习更多
阅读 [FP Complete](https://www.fpcomplete.com/rust/) 在 Rust 方面的更多信息，并看看我们的其他学习材料。如果你有兴趣在你的项目中获得帮助，请查看我们提供的[咨询](https://www.fpcomplete.com/consulting-services/)和[培训](https://www.fpcomplete.com/training/)服务。

FP Complete 专注于服务器端软件，具有 Rust、 [Haskell](https://www.fpcomplete.com/haskell/) 和 [DevOps](https://www.fpcomplete.com/devsecops/) 方面的专业知识。如果您有兴趣了解我们如何帮助您的团队取得成功，请与我们的工程师之一进行[免费咨询](https://www.fpcomplete.com/contact-us/)。
