>* Hecto: Build your own text editor in Rust 译文（用 Rust 构建自己的文本编辑器）
>* 原文链接：https://www.philippflenker.com/hecto-chapter-2/
>* 原文作者：[Philipp Flenker](https://github.com/pflenker)
>* 译文来自：https://github.com/suhanyujie/article-transfer-rs/
>* 译者：[suhanyujie](https://ishenghuo.cnblogs.com/)
>* ps：水平有限，如有不当之处，欢迎指正
>* 标签：Rust，text-editor

接下来我们尝试从用户那里读取按键输入。删除默认的 main 函数，修改为如下:

```rust
use std::io::{self, Read};

fn main() {
    for b in io::stdin().bytes() {
        let c = b.unwrap() as char;
        println!("{}", c);
    }
}
```

也可以参考 [github](https://github.com/pflenker/hecto-tutorial/tree/read-keys) 上的代码。

你可以试试这个程序，看看它是如何工作的。要停止它，按 Ctrl-C。

首先，我们使用 use 将依赖库导入到我们的程序中。我们想对系统的输入/输出进行一些操作，简而言之就是 io。所以我们使用 `use std::io::{self, Read}`，它为我们导入 io 库:

```rust
use std::io;
use std::io::Read;
```

在此之后，我们能够在代码中使用 io 库中的代码，并且将 `Read` 引入到代码中使我们能够使用 `byte()`。尝试在不导入 `Read` 的情况下运行代码，编译器将出现一个错误，解释 `Read` 实际上需要在作用域中，因为它提供了 `bytes` 的实现。这个概念被称为 trait，我们将在本教程的后面更深入地研究 trait。关于 [trait 的文档](https://doc.rust-lang.org/book/ch10-02-traits.html)你可以添加到你的阅读列表中！

如果你是 Rust 的新手，别担心。在本章中我们还要学习一些东西，但是以后添加的代码不会像这一章一下子带来那么多新概念。另外，随着教程的深入，一些概念会变得更加清晰，所以如果您不能一下子理解所有内容，也不必担心。

`main` 中的第一行执行了许多操作，可以概括为“对于每次从键盘读取的字节，将其绑定到 b 变量上并执行后续代码块”。

让我们现在解析 line。`io::stdin()` 意味着我们想从来自于 io 库中的名为 stdin 的方法调用。io 是我们刚刚导入的内容。`stdin` 表示[标准输入流](https://en.wikipedia.org/wiki/Standard_streams#Standard_input_(stdin))，简单地说，它使您能够访问输入到程序中的所有参数。

在 `io::stdin()` 上调用 `byte()` 返回我们可以迭代的内容，或者换句话说: 让我们在一系列元素上执行相同任务的内容。与许多其他语言一样，在 Rust 中，这个概念被称为[迭代器](https://doc.rust-lang.org/book/ch13-02-iterators.html)。

使用迭代器可以让我们用 `for..in` 构建一个循环。`for..in` 为什么与 `byte()` 结合 ，我们要求 Rust 将字节从标准输入读取到变量 b，并持续这样做，直到没有字节可以读取。`for..in` 后面的两行打印出每个字符 —— 我们将稍后解释 `unwrap` 和 `println!`，如果没有读取的数据，则直接返回。

当你执行 `./hecto` 时，您的终端连接到标准输入，因此您的键盘输入被读入 b 变量。但是，默认情况下，终端以规范模式（canonical mode）启动，也称为熟模式（cooked mode）。在这种模式下，键盘输入只有在用户按下 Enter 时才会发送到程序。这对许多程序都很有用: 它让用户输入一行文本，使用 Backspace 进行修改，直到他们得到他们想要的输入，最后按 Enter 发送到程序。但是，对于具有更复杂用户界面的程序(如文本编辑器)来说，它并不能很好地工作。我们想处理每一个按键，因为一旦按钮被按下，我们我们需要立即作出反应。

要退出上面的程序，按 Ctrl-D 告诉 Rust 它已经到达文件的末尾。或者您总是可以按 Ctrl-C 来向进程发出立即终止的信号。

我们要的是原始模式（raw mode）。幸运的是，有外部库可以将终端设置为原始模式。Rust 中的图书馆被称为“板条箱（crate）” —— 如果你想了解更多，这里有[文档的链接](https://doc.rust-lang.org/book/ch07-00-managing-growing-projects-with-packages-crates-and-modules.html)。像许多其他编程语言一样，Rust 有一个精简的核心，并依赖于板条箱来扩展其功能。在本教程中，我们有时会在切换到外部功能之前首先手动完成一些操作，有时我们会直接使用库函数。

## 按 Q 退出
为了演示 canonical 模式是如何工作的，我们将让程序在从用户那里读取 q 键按下时退出。

```rust
use std::io::{self, Read};
fn main() {
    for b in io::stdin().bytes() {
        let c = b.unwrap() as char;
        println!("{}", c);
        if c == 'q' {
            break;
        }
    }
}
```

[查看 GitHhub 源码](https://github.com/pflenker/hecto-tutorial/tree/q-to-quit)

请注意，在 Rust 中，字符声明需要单引号 `'`，而不是双引号 `"`！

要退出该程序，您必须键入一行包含 `q` 的文本，然后按 Enter。程序将快速读取文本行，每次读取一个字符，直到 `q`，此时 `for..in` 循环将停止，程序将退出。`q` 之后的任何字符都将留在输入队列上不读，也不会打印出来。Rust 在程序退出时将丢弃它们。

## 使用 Termion 进入原始模式
更改您的 Cargo.toml，追加一个依赖，如下:

```
termion = "1"
```

这样，我们就告诉 cargo，我们有一个版本为 1 的称为 termion 的依赖项。Cargo 遵循一个名为 [Semantic Versioning](https://semver.org/) 的概念，其中一个程序版本通常由三个数字组成(比如 `0.1.0`) ，按照惯例，只要第一个数字保持不变，就不会发生太大变化。这意味着，如果您针对版本 `v1.5.0` 进行开发，那么您的程序也可以使用版本号为 `v1.5.1`，甚至版本号为 `v1.7.0` 的依赖库。这很有用，因为这意味着我们可以获得带有 bug 修复和小的新特性的依赖库，但是现有的特性仍然可以使用，而不需要我们修改代码。通过设置 `termion = "1`"，我们可以获取并使用以 1 开头的最新版本。

下次您运行 `cargo build`  或 `cargo run`，新的依赖项，将被下载和编译，输出将看起来像这样：

```
   Compiling libc v0.2.62
   Compiling numtoa v0.1.0
   Compiling termion v1.5.3
   Compiling hecto v0.1.0 (/home/philipp/repositories/hecto)
    Finished dev [unoptimized + debuginfo] target(s) in 7.83s
```

正如您所看到的，`termion` 本身带有很多依赖项，通过 `cargo` 下载和编译。您可能注意到 Cargo.lock 也发生了变化: 它现在包含了所有已安装的包和依赖项的确切名称和版本。这有助于避免“在我的机器上出问题” —— 如果你在一个团队中工作，当你遇到一个 bug 时，比如说在 1.2.3 版本中，而你的同事在 1.2.4 版本中却没有看到它。

> 如果你在第一章没有看到，这个教程不能在 Windows 上完成。Termion 不支持该平台，但是您可以使用 windows 的  Linux 子系统运行它。在编写本教程时，Termion 是本教程的唯一选项。现在，在本教程发布两年之后，还有一个跨平台库可用: cross-sterm。它的工作原理与 Termion 不同，所以如果你选择使用 Windows 而不是 Linux 系统，你不能直接按照这个教程学习，但是如果你已经有软件编程经验，你可能会发现尝试使用它来代替 Termion 很容易，也很有吸引力。
> 如果你正在寻找交叉编译的方案，使用这个令人敬畏的 hecto 类似的东西可以在所有平台上运行。

现在将 main.rs 修改如下:

```rust
use std::io::{self, stdout, Read};
use termion::raw::IntoRawMode;

fn main() {
    let _stdout = stdout().into_raw_mode().unwrap();

    for b in io::stdin().bytes() {
        let c = b.unwrap() as char;
        println!("{}", c);
        if c == 'q' {
            break;
        }
    }
}
```

[参考 GitHub](https://github.com/pflenker/hecto-tutorial/tree/into-raw-mode)

尝试一下，您会发现您输入的每个字符都会立即打印出来，一旦您输入 `q`，程序就结束了。

发生什么了？

这里有几点需要注意。首先，我们使用 `termion` 来提供标准输出(stdout) ，它是前面提到的标准输入的对应的东西，带有一个名为 `into_raw_mode()` 的函数，我们正在调用该函数。但是为什么要在标准输出上调用这个方法来改变我们从标准输入读取的数据呢？答案是，终端的状态由写入器控制，而不是由读取器控制。writer 用于在屏幕上绘制或移动光标，因此它也用于更改模式。

其次，我们将 into_raw_mode 的结果赋给一个名为 `_stdout` 的变量，但是我们不对该变量进行任何操作。为什么？因为这是我们第一次遇到 Rust 的[所有权系统](https://doc.rust-lang.org/book/ch04-00-understanding-ownership.html)。概括一个复杂的概念，函数可以拥有某些东西。但不属于自己的东西会被移走。`into_raw_mode` 修改终端并返回一个值，这个值一旦被删除，就会将终端重置为 canonical 模式 —— 所以我们需要通过将其绑定到 `_stdout` 来保留它。可以通过删除 `let _stdout =` —— 此时终端不会保持原始模式，进行尝试。


