>* Parsing bitstreams with Nom 译文（用 Nom 解析比特流）
>* 原文链接：https://blog.adamchalmers.com/nom-bits/
>* 原文作者：[Adam Chalmers](https://twitter.com/adam_chal)
>* 译文来自：https://github.com/suhanyujie/article-transfer-rs/
>* 译者：[suhanyujie](https://github.com/suhanyujie)
>* ps：水平有限，翻译不当之处，还请指正。
>* 标签：Rust, Nom, parse, bit stream

编程语言通常只处理字节（8 位一组）。操作单个位可能非常棘手。但有时你的确需要这样做，例如，[DNS 头部](https://datatracker.ietf.org/doc/html/rfc1035#section-4.1.1)有一些 4 比特位的值，而且会将一些布尔值编码为单个位。所以我们真的需要一种方法来解析二进制数据，而不是将所有都数据都解析w为 8 位的字节处理。

幸运的是，[Nom](https://docs.rs/nom) 可以做到这一点！在上一篇博文中，我们学习了如何使用 Nom [解析文本文件](https://blog.adamchalmers.com/nom-chars)，方法是从只解析几个字符的简单解析器开始。然后，使用组合子，将这些简单的解析器组合成更复杂的可以反序列化整个结构化文件的解析器。我们也可以复用这种方法来解析二进制数据。让我们来看看!

注意，所有的代码示例都使用了 Nom 7 —— 如果 Nom 8 版本做出了破坏性的改变，我会尝试更新它，但我不能保证 :)

## 表示比特流输入

在前一篇文章中，我们看到 Nom 解析器包含三种类型的泛型：

* `I` 表示输入
* `O` 表示输出
* `E` 代表错误类型

在上一篇文章中，我们使用了 `I = &str`，它可以让你解析文本流。输入类型 `I = &[u8]` 允许你解析字节流。但是我们如何表示一个比特流呢? Rust 没有任何类型来表示它（比特位）!

这个问题实际上非常重要，在这篇博文中会多次提到。我们如何表示位数 n < 8 的比特位?

我首先想到的是“用bool” —— 你可能也在这么想。只需将 0 表示为 `false`，将 1 表示为 `true`。这肯定有效，但有点浪费。Rust 的 bool 类型实际上占用一个字节。因此，你可以用 `&[bool]` 表示比特比特流 ，但它将占用所需空间的 8 倍多字节！这对于一些应用程序来说都是可以，但是 Nom 可以更高效。

The solution to representing a bitstream is pretty clever: we use two values. Firstly, a list of bytes, and secondly, a number representing which exact bit of those bytes is next.
> 表示比特流（bitstream）的解决方案很巧妙：我们使用两个值。首先是一个字节列表，其次是一个数字，表示下一个字节的具体位置。

```rust
I = (&[u8], usize)
```

这对值表示一个比特流（bitstream）。第一个元素是比特流数据，每次存储一个字节（所以，8位一组）。第二个元素是偏移量，显示接下来应该读取哪一位（不是字节！）

例如，假设我们想要解析一个 16 位的序列，比如 `1111 0000 1100 1100`。我们从第一个元素位开始。下面展示了我如何表示比特流和偏移量：

```
1111000011001100
^
```

注意，这个小箭头指向输入的第一个比特位。为了使用 `I = (&[u8], usize)` 类型表示这一点，我们将比特位输入分解为字节，以及当前指向的位的索引。如下所示：

```
([0b11110000, 0b11001100], 0)
    ^
```

假设你从那里解析了 3 位。比特流将的表示是

```
([0b11110000, 0b11001100], 3)
       ^
```

再读 6 位后，输入的第一个字节就会读完：

```
([0b11110000, 0b11001100], 9)
                 ^
```

因为第一个字节将永远不会再次使用，所以 Nom 通过删除第一个字节进行优化，并调整偏移量来实现这一点:

```
 ([0b11001100], 1)
      ^
```

因为这个元组类型出现得太频繁了，我通常会为它添加一个类型别名:

```rust
type BitInput<'a> = (&'a [u8], usize);
```

## Parsing bitstreams with "take"
我们刚刚学习了 Nom 如何表示处理的单个比特。现在我们来看看如何在比特流输入上执行解析器的逻辑。有两个主要的位级别的解析器：`take` 和 `tag`。

[nom::bits::complete::take](https://docs.rs/nom/7.1.0/nom/bits/complete/fn.take.html) 解析器类似于上一篇文章中的 [nom::bytes::complete::take](https://docs.rs/nom/7.1.0/nom/bytes/complete/fn.take.html) 解析器。它有一个参数 count，表示从输入中获取的位数，然后返回这些 `count` 数量的比特位作为输出。输入中如果没有足够的位剩余（如：即输入的位长度 < `count`)，那么它会 panic。

这个解析器听起来很简单，但是调用它需要一点工作，因为它有很多泛型[1]，而 Rust 通常不能推断出这个类型。我总是提供特定具体类型的辅助函数。那么，让我们构建一个解析四个比特的解析器。一个 4 位的数字被称为 _nibble_ （因为它是半个字节……以工程标准来看...得到它...，这简直就是个“笑话”）。

显然，它的 I（通用输入类型）将是标准的 `BitInput` 别名。但是 O（通用输出类型）是什么呢？毕竟，Rust 没有 u4 类型。我们如何存储一个 4 个比特位的数字？

这个问题实际上只是早期的问题，即“我们如何在 Rust 中表示比特位”。瞧，我告诉过你在这篇文章里它出现了好几次。

而 Nom 的 `take` 解析器通过在 n 位中填充前导零来解决这一问题，并将它们存储在某种 uint 类型中，如 `u8`、`u16` 或其他类型。我们将把前 4 位解析成 0~15 之间的一个数字，然后把这个数字存储在 u8 中。这意味着高 4 位将永远为零。你可以使用任意的 uint 类型，但是使用 `u8` 已经绰绰有余了，并且比使用 `u16` 或任何更大的类型浪费更少的内存。

```rust
use nom::{bits::complete::take, IResult};
type BitInput<'a> = (&'a [u8], usize);

/// 从 BitInput 中取走 4 个比特位
/// 将输出保存在 u8 类型中, 由于没有 u4 类型，因而 u8 是最合适的
fn take_nibble(i: BitInput) -> IResult<BitInput, u8> {
    // Have to specify some concrete numeric type, otherwise Rust won't know which
    // type of number you're trying to use here. I used usize, but you could use
    // any uint type.
    take(4usize)(i)
}
// Note that Rust number literals let you put underscores wherever you'd like, to
// enhance readability. E.g. you can help separate commas, by writing 1000000 as
// 1_000_000.
// I've used them here to visually separate the two u4 values in this u8.
let input = ([0b1010_1111].as_ref(), 0);

let (_input, actual_nibble) = take_nibble(input).unwrap();
let expected_nibble = 0b1010;
assert_eq!(actual_nibble, expected_nibble);
```

## 用 "tag" 解析比特流

按位解析的 [tag](https://docs.rs/nom/7.1.0/nom/bits/complete/fn.tag.html) 解析器匹配特定的位模型，如“0110”。同样，这是一个简单的想法，但它提出了一个棘手的问题：我们如何表示位模型？是的，这是我第三次问“Rust 中我们如何表示比特位”。我告诉过你，这是一个很基本的问题！这里使用了一个非常相似的解决方案。

Nom represents the bit pattern using two parameters:
> Nom 使用两个参数表示位模型：

* **count**: 样本数据有多少位
* **pattern**: 样本，前导以零填充，以拼接成适合某个 uint 类型。

比如：

* 样本 101 表示为（样本：00000_101，计数：3）。
* 样本 111000111 表示为（样本：0000000_111000111，计数：9）。

作为程序员，你将选择对 `pattern` 和 `count` 的存储使用哪种 uint 类型 -- 解析器对各种 uint 类型都通用。我个人只会使用适合该值的最小单位。因此，示例 1 的样本适合 `u8`，示例 2 的样本适合于 u16，而在这两种情况下，计数都适用 `u8`。所以我会按照这样的原则使用。不过我觉得这不是很重要。

好了，现在我们知道了如何表示位数据，tag 解析器就很容易解析了。你提供一个位数据，Nom 逐位将其与输入位进行比较。像所有的解析器一样，tag 返回一个结果。如果解析器与输入匹配，则返回 Ok 和一对数据：（剩余的 input，匹配的 output）。如果没有匹配项，则返回 Err。因此，假如，解析比特流 10100 上的样本 101 时，将返回 `(00, 101)`。

现在，我们准备看实际代码：

```rust
use nom::{bits::complete::tag, IResult};

type BitInput<'a> = (&'a [u8], usize);

// This is just a simple wrapper around the `tag` parser, but it makes the
// parameter types concrete instead of generic, so now Rust knows how to actually
// store the pattern.
fn parser(pattern: u8, count: u8, input: BitInput) -> IResult<BitInput, u8> {
    tag(pattern, num_bits_to_compare)(input)
}

// The pattern 1111 matches the stream 1111_1111
assert!(parser(0b1111, 4, (&[0b1111_1111], 0)).is_ok());
// The pattern 1 matches the stream too
assert!(parser(0b1, 1, (&[0b1111_1111], 0)).is_ok());
// The pattern 01 does _not_ match the stream
assert!(parser(0b1, 2, (&[0b1111_1111], 0)).is_err());
// The pattern 1111_1110 doesn't match the stream either
assert!(parser(0b1111_1110, 8, (&[0b1111_1111], 0)).is_err());
```

## 比特流和组合子（组合器）

记住，解析器组合器库的整个思想是：从几个小解析器开始，然后将它们与组合器函数组合在一起。我们已经看到了两个基本的“比特流”解析器，tag 和 take。以下是如何将它们与我上一篇帖子中的 [map](https://docs.rs/nom/7.1.0/nom/combinator/fn.map.html) 组合器相结合：

```rust
use nom::{bits::complete::take, combinator::map, IResult};
type BitInput<'a> = (&'a [u8], usize);

/// Takes one bit from the input, returning true for 1 and false for 0.
fn take_bit(i: BitInput) -> IResult<BitInput, bool> {
    map(take(1usize), |bits: u8| bits > 0)(i)
}

let input = ([0b10101010].as_ref(), 0);
let (input, first_bit) = take_bit(input).unwrap();
assert!(first_bit); // First bit is 1
let (_input, second_bit) = take_bit(input).unwrap();
assert!(!second_bit); // Second bit is 0
```

## Converting bytestreams to bitstreams and back

到目前为止，我们已经学习了如何制作简单的位解析器，并将它们组合成复杂的解析器。我们甚至已经了解了 Nom 如何表示比特流。最后一个问题是：这些比特流到底是从哪里来的？毕竟，大多数 Rust 函数以字节表示二进制数据（例如，`Vec<u8>` 或使用 bytes crate）。如果你从磁盘、RAM 或网络读取二进制数据，那么几乎可以肯定它将以字节存储。因此，我们需要一种将字节流转换为比特流的方法。

幸运的是，函数 `nom::bits::bits` 就是做这件事的。文档说明了，它“将字节级输入转换为位级输入，供使用位的解析器使用。”太棒了！

同样，该函数使用了很多可能会令人困惑的泛型，因此这里有一个示例说明它是如何工作的。

```rust
use nom::IResult;
use nom::multi::many0;
use nom::number::complete::be_u16;

type BitInput<'a> = (&'a [u8], usize);

/// Stub example type. Imagine this has to be parsed from individual bits.
struct BitwiseHeader;

/// A bit-level parser
fn parse_header(i: BitInput) -> IResult<BitInput, BitwiseHeader> {
    todo!()
}

/// Stub example type.
/// The header has to be parsed from bits, but the body can be parsed from bytes.
struct Message {
    header: BitwiseHeader,
    body: Vec<u16>,
}

/// A byte-level parser that calls a bit-level parser
fn parse_msg(i: &[u8]) -> IResult<&[u8], Message> {
    /// The header has to be parsed from bits
    let (i, header) = nom::bits::bits(parse_header)(i)?;
    /// But the rest of the message can be parsed from bytes.
    let (i, body) = many0(be_u16)(i)?;
    Ok((i, Message { header, body }))
}
```

因为 [Advent of Code 2021，day 16](https://adventofcode.com/2021/day/16) 的缘故，我对这些位解析器感到好奇。几周后，我想[构建一个 DNS 客户端](https://github.com/adamchalmers/dingo/blob/d1a34e37b2c743dcd63bfe8612fd1d7c63ce9d63/src/message.rs#L239)，并意识到使用 Nom 解析 u4 或单比特位很容易。在我的下一篇博客文章中，我们将研究真实场景中的如何解析比特流：DNS 消息头。

1.看看它的[类型签名](https://docs.rs/nom/7.1.0/nom/bits/complete/fn.take.html)。概括地说，如下所示：

* `input`: I 必须是类似于二进制流的数据
* `count`: C 必须是某个 uint 类型（带有[一些限定](https://docs.rs/nom/7.1.0/nom/trait.ToUsize.html)）
* `output`: `O` 必须是某种 uint 类型，但具有不同的限定
* 当然，对于详细的解析器错误通常使用 `E` 表示，也可以使用帮助较小的默认解析器异常表示。
